This is the mail archive of the gdb-patches@sourceware.org mailing list for the GDB project.


Index Nav: [Date Index] [Subject Index] [Author Index] [Thread Index]
Message Nav: [Date Prev] [Date Next] [Thread Prev] [Thread Next]
Other format: [Raw text]

non-stop linux native, unbreak -exec-interrupt --all, add an MI test


Hi,

This patch unbreaks "interrupt -a" in linux native, non-stop mode.

The issue here is that we currently are using SIGINT to interrupt
all lwps, by sending each one a signal, and this goes against
the logic in place to filter out SIGINT's when they are delivered
to each lwp in the process group.

I've also added a new MI testcase to explicitly test that
the -exec-interrupt --all works as expected.

While doing that, I noticed that mi_expect_stop would sometimes
match the  wrong thing or eat to much input if input
came intermingled with other notifications and warnings.

This also triggers sometimes in the existing mi-nonstop.exp
test.  E.g., if I run it sufficient times, I end up seeing
this failure:

-exec-continue
^running
*running,thread-id="1"
(gdb) 
=thread-created,id="2"
~"[New Thread 0x55923b90 (LWP 1174)]\n"
*running,thread-id="2"
mi_expect_stop: expecting: 
\*stopped,reason="breakpoint-hit",disp="keep",bkptno="[0-9]+",thread-id="[0-9]+",stopped-threads=.*,frame={addr="0x[0-9A-Fa-f]+",func="break_at_me",args=\[.*\],file=".*non-stop.c",fullname="(/.*/|
\\\\[^\\]+\\.+\\|\\[^\\].*\\|[a-zA-Z]:.*\\)non-stop.c",line=".*"}
$
*stopped,reason="breakpoint-hit",disp="keep",bkptno="2",thread-id="2",stopped-threads=["2"],frame={addr="0x08048507",func="break_at_me",args=[{name="id",value="0"},
{name="i",value="0"}],file="../../../src/gdb/testsuite/gdb.mi/non-stop.c",fullname="/home/pedro/gdb/baseline/src/gdb/testsuite/gdb.mi/non-stop.c",line="40"}
=thread-created,id="3"
~"[New Thread 0x55b24b90 (LWP 1175)]\n"
*running,thread-id="3"
got =thread-created,id="2"
~"[New Thread 0x55923b90 (LWP 1174)]\n"
*running,thread-id="2"
*stopped,reason="breakpoint-hit",disp="keep",bkptno="2",thread-id="2",stopped-threads=["2"],frame={addr="0x08048507",func="break_at_me",args=[{name="id",value="0"},
{name="i",value="0"}],file="../../../src/gdb/testsuite/gdb.mi/non-stop.c",fullname="/home/pedro/gdb/baseline/src/gdb/testsuite/gdb.mi/non-stop.c",line="40"}
=thread-created,id="3"
~"[New Thread 0x55b24b90 (LWP 1175)]\n"
*running,thread-id="3"

FAIL: gdb.mi/mi-nonstop.exp: w0,i0 stop (stopped at wrong place)

The "pass" regex fails to match the input because it is relying on
the *stopped notification to show at the the end of input, but sometimes the
OS provides enough input after the *stopped line to make it fail to match.

The "(stopped at wrong place)" regex ends with a ".*\r\n$", so it
matches this case.  Note that in async mode, there is no gdb
prompt to expect at the end of a notification.

I've fixed it by removing the trailing anchor and making the regex match
all input before the *stopped notification non-greedily, so
it matches on the first *stopped notification it sees, and only consumes
one notification.  Perhaps there another way to do this?

-- 
Pedro Alves
gdb/
2008-08-22  Pedro Alves  <pedro@codesourcery.com>

	* linux-nat.c (linux_nat_wait): Only use set_ignore_sigint in
	all-stop mode.

gdb/testsuite/
2008-08-22  Pedro Alves  <pedro@codesourcery.com>

	* gdb.mi/mi-nsintrall.exp: New.
	* gdb.mi/intrall.c: New.
	* lib/mi-support.exp (mi_expect_stop): Only consume one *stopped
	notification.
	(mi_expect_interrupt): New.

---
 gdb/linux-nat.c                       |   26 +++--
 gdb/testsuite/gdb.mi/intrall.c        |   64 ++++++++++++++
 gdb/testsuite/gdb.mi/mi-nsintrall.exp |  155 ++++++++++++++++++++++++++++++++++
 gdb/testsuite/lib/mi-support.exp      |   42 ++++++++-
 4 files changed, 273 insertions(+), 14 deletions(-)

Index: src/gdb/testsuite/gdb.mi/mi-nsintrall.exp
===================================================================
--- /dev/null	1970-01-01 00:00:00.000000000 +0000
+++ src/gdb/testsuite/gdb.mi/mi-nsintrall.exp	2008-08-22 19:34:51.000000000 +0100
@@ -0,0 +1,155 @@
+# Copyright 2008 Free Software Foundation, Inc.
+
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 3 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+
+# This only works with native configurations
+if {![isnative]} {
+  return
+}
+
+load_lib mi-support.exp
+set MIFLAGS "-i=mi"
+
+gdb_exit
+if {[mi_gdb_start]} {
+    continue
+}
+
+proc myreverse { list } {
+    if { [llength $list] <= 1 } {
+        return $list
+    }
+    set tail [lrange $list 1 [llength $list]]
+    set rtail [myreverse $tail]
+    lappend rtail [lindex $list 0]
+    return $rtail
+}
+
+proc check_thread_states { xstates test } {
+    global expect_out
+    set states [myreverse $xstates]
+    set pattern "\\^done,threads=\\\["
+    foreach s $states {
+        set pattern "${pattern}(.*)state=\"$s\""
+    }
+    set pattern "$pattern\\\}\\\].*"
+
+    verbose -log "expecting: $pattern"
+    mi_gdb_test "-thread-info" $pattern $test
+}
+
+#
+# Start here
+#
+set testfile "intrall"
+set srcfile "$testfile.c"
+set binfile "$objdir/$subdir/mi-$testfile"
+
+set options [list debug incdir=$objdir]
+if {[gdb_compile_pthreads "$srcdir/$subdir/$srcfile" $binfile executable $options] != "" } {
+    return -1
+}
+
+mi_gdb_reinitialize_dir $srcdir/$subdir
+mi_gdb_load $binfile
+
+set supported 0
+send_gdb "-gdb-show non-stop\n"
+gdb_expect {
+    -re ".*\\^done,value=\"off\",supported=\"(\[^\"\]+)\"\r\n$mi_gdb_prompt$" {
+        if { $expect_out(1,string) == "1" } {
+            set supported 1
+        }
+    }
+    -re ".$mi_gdb_prompt$" {
+    }
+}
+
+mi_gdb_test "-gdb-set non-stop 1" ".*"
+mi_gdb_test "-gdb-set target-async 1" ".*"
+detect_async
+
+mi_gdb_test "200-break-insert -t main" ".*"
+# Note: presently, we skip this test on non-native targets,
+# so 'run' is OK.  As soon as we start to run this on remote
+# target, the logic from mi_run_cmd will have to be refactored.
+send_gdb "-exec-run\n"
+gdb_expect {
+    -re ".*\\^running.*$mi_gdb_prompt$" {
+    }
+    -re "\\^error,msg=\"The target does not support running in non-stop mode.\"" {
+        verbose -log "Non-stop mode not supported, skipping all tests"
+        return
+    }
+    -re ".*\r\n$mi_gdb_prompt$" {
+        perror "Cannot start target (unknown output after running)"
+        return -1
+    }
+    timeout {
+        perror "Cannot start target (timeout)"
+        return -1
+    }
+}
+mi_expect_stop "breakpoint-hit" main ".*" ".*" "\[0-9\]+" { "" "disp=\"del\"" } "run to main"
+
+mi_create_breakpoint thread_function 2 keep thread_function .* .* .* "breakpoint at thread_function"
+
+mi_send_resuming_command "exec-continue --all" "resume all"
+for {set i 0} {$i < 6} {incr i} {
+    mi_expect_stop "breakpoint-hit" "thread_function" ".*" "intrall.c" ".*" {"" "disp=\"keep\""} "stop $i"
+}
+
+check_thread_states {"stopped" "stopped" "stopped" "stopped" "stopped" "stopped"} "thread state, all stopped"
+
+# now that we know about all the threads, we can get rid of the breakpoints
+mi_delete_breakpoints
+
+set running_re ""
+for {set i 6} {$i > 0} {incr i -1} {
+    set running_re "$running_re\\*running,thread-id=\"$i\"\r\n"
+}
+
+send_gdb "-exec-continue --all\n"
+gdb_expect {
+    -re ".*$running_re$mi_gdb_prompt" {
+        pass "resume all, no breakpoint"
+    }
+    timeout {
+        fail "resume all, no breakpoint (timeout)"
+    }
+}
+check_thread_states {"running" "running" "running" "running" "running" "running"} "thread state, resume all"
+
+send_gdb "-exec-interrupt --all\n"
+for {set i 0} {$i < 6} {incr i} {
+    mi_expect_interrupt "interrupt $i"
+}
+check_thread_states {"stopped" "stopped" "stopped" "stopped" "stopped" "stopped"} "thread state, all interrupted"
+
+# resume all threads, and check that no thread is spuriously left interrupted
+send_gdb "-exec-continue --all\n"
+gdb_expect {
+    -re ".*$running_re$mi_gdb_prompt" {
+        pass "resume all after interrupting"
+    }
+    timeout {
+        fail "resume all after interrupting (timeout)"
+    }
+}
+
+sleep 1
+check_thread_states {"running" "running" "running" "running" "running" "running"} "thread state, resume all after interrupting"
+
+mi_gdb_exit
+
Index: src/gdb/testsuite/gdb.mi/intrall.c
===================================================================
--- /dev/null	1970-01-01 00:00:00.000000000 +0000
+++ src/gdb/testsuite/gdb.mi/intrall.c	2008-08-22 19:34:51.000000000 +0100
@@ -0,0 +1,64 @@
+/* This testcase is part of GDB, the GNU debugger.
+
+   Copyright 2002, 2003, 2004, 2007, 2008 Free Software Foundation, Inc.
+
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 3 of the License, or
+   (at your option) any later version.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
+
+#include <stdio.h>
+#include <unistd.h>
+#include <stdlib.h>
+#include <pthread.h>
+
+void *thread_function(void *arg); /* Pointer to function executed by each thread */
+
+#define NUM 5
+
+unsigned int args[NUM+1];
+
+int main() {
+    int res;
+    pthread_t threads[NUM];
+    void *thread_result;
+    long i;
+
+    for (i = 1; i <= NUM; i++)
+      {
+	args[i] = 1;
+	res = pthread_create(&threads[i],
+		             NULL,
+			     thread_function,
+			     (void *) i);
+      }
+
+    /* last thread start.  */
+    args[0] = 1;
+    thread_function ((void *) 0);
+
+    exit(EXIT_SUCCESS);
+}
+
+void *thread_function(void *arg) {
+    int my_number =  (long) arg;
+    int *myp = (int *) &args[my_number];
+
+    /* Don't run forever.  Run just short of it :)  */
+    while (*myp > 0)
+      {
+	/* main loop.  */
+	(*myp) ++;
+	usleep (1);
+      }
+
+    pthread_exit(NULL);
+}
Index: src/gdb/testsuite/lib/mi-support.exp
===================================================================
--- src.orig/gdb/testsuite/lib/mi-support.exp	2008-08-22 19:27:28.000000000 +0100
+++ src/gdb/testsuite/lib/mi-support.exp	2008-08-22 19:34:51.000000000 +0100
@@ -1009,13 +1009,13 @@ proc mi_expect_stop { reason func args f
 
     set a $after_reason
 
-    verbose -log "mi_expect_stop: expecting: \\*stopped,${r}${a}${bn}thread-id=\"$decimal\",stopped-threads=.*,frame=\{addr=\"$hex\",func=\"$func\",args=$args,file=\".*$file\",fullname=\"${fullname_syntax}$file\",line=\"$line\"\}\r\n$after_stopped$prompt_re$"
+    verbose -log "mi_expect_stop: expecting: \\*stopped,${r}${a}${bn}thread-id=\"$decimal\",stopped-threads=.*,frame=\{addr=\"$hex\",func=\"$func\",args=$args,file=\".*$file\",fullname=\"${fullname_syntax}$file\",line=\"$line\"\}\r\n$after_stopped$prompt_re"
     gdb_expect {
-	-re "\\*stopped,${r}${a}${bn}thread-id=\"$decimal\",stopped-threads=.*,frame=\{addr=\"$hex\",func=\"$func\",args=$args,file=\".*$file\",fullname=\"${fullname_syntax}$file\",line=\"($line)\"\}$after_stopped\r\n$prompt_re$" {
+	-re ".*?\\*stopped,${r}${a}${bn}thread-id=\"$decimal\",stopped-threads=.*,frame=\{addr=\"$hex\",func=\"$func\",args=$args,file=\".*$file\",fullname=\"${fullname_syntax}$file\",line=\"($line)\"\}$after_stopped\r\n$prompt_re" {
 	    pass "$test"
-            return $expect_out(2,string)
+	    return $expect_out(2,string)
 	}
-	-re "\\*stopped,${r}${a}${bn}thread-id=\"$decimal\",stopped-threads=.*,frame=\{addr=\"$hex\",func=\".*\",args=\[\\\[\{\].*\[\\\]\}\],file=\".*\",fullname=\"${fullname_syntax}.*\",line=\"\[0-9\]*\"\}.*\r\n$prompt_re$" {
+	-re ".*?\\*stopped,${r}${a}${bn}thread-id=\"$decimal\",stopped-threads=.*,frame=\{addr=\"$hex\",func=\".*\",args=\[\\\[\{\].*\[\\\]\}\],file=\".*\",fullname=\"${fullname_syntax}.*\",line=\"\[0-9\]*\"\}.*\r\n$prompt_re" {
             verbose -log "got $expect_out(buffer)"
 	    fail "$test (stopped at wrong place)"
 	    return -1
@@ -1032,6 +1032,40 @@ proc mi_expect_stop { reason func args f
     }    
 }
 
+# Wait for MI *stopped notification related to an interrupt request to
+# appear.
+proc mi_expect_interrupt { test } {
+    global mi_gdb_prompt
+    global decimal
+    global async
+
+    if {$async} {
+        set prompt_re ""
+    } else {
+        set prompt_re "$mi_gdb_prompt"
+    }
+
+    set r "reason=\"signal-received\",signal-name=\"SIGINT\",signal-meaning=\"Interrupt\""
+
+    # A signal can land anywhere, just ignore the location
+    verbose -log "mi_expect_interrupt: expecting: \\*stopped,${r},thread-id=\"$decimal\",stopped-threads=.*.*\r\n$prompt_re"
+    gdb_expect {
+	-re ".*?\\*stopped,${r},thread-id=\"$decimal\",stopped-threads=.*\r\n$prompt_re" {
+	    pass "$test"
+	    return 0;
+	}
+	-re ".*\r\n$mi_gdb_prompt$" {
+            verbose -log "got $expect_out(buffer)"
+	    fail "$test (unknown output after running)"
+	    return -1
+	}
+	timeout {
+	    fail "$test (timeout)"
+	    return -1
+	}
+    }
+}
+
 # cmd should not include the number or newline (i.e. "exec-step 3", not
 # "220-exec-step 3\n"
 

Index Nav: [Date Index] [Subject Index] [Author Index] [Thread Index]
Message Nav: [Date Prev] [Date Next] [Thread Prev] [Thread Next]
Index: src/gdb/linux-nat.c
===================================================================
--- src.orig/gdb/linux-nat.c	2008-08-22 19:27:28.000000000 +0100
+++ src/gdb/linux-nat.c	2008-08-22 19:34:51.000000000 +0100
@@ -2935,18 +2935,24 @@ retry:
 	  goto retry;
 	}
 
-      if (signo == TARGET_SIGNAL_INT && signal_pass_state (signo) == 0)
+      if (!non_stop)
 	{
-	  /* If ^C/BREAK is typed at the tty/console, SIGINT gets
-	     forwarded to the entire process group, that is, all LWPs
-	     will receive it - unless they're using CLONE_THREAD to
-	     share signals.  Since we only want to report it once, we
-	     mark it as ignored for all LWPs except this one.  */
-	  iterate_over_lwps (set_ignore_sigint, NULL);
-	  lp->ignore_sigint = 0;
+	  /* Only do the below in all-stop, as we currently use SIGINT
+	     to implement target_stop (see linux_nat_stop) in
+	     non-stop.  */
+	  if (signo == TARGET_SIGNAL_INT && signal_pass_state (signo) == 0)
+	    {
+	      /* If ^C/BREAK is typed at the tty/console, SIGINT gets
+		 forwarded to the entire process group, that is, all LWPs
+		 will receive it - unless they're using CLONE_THREAD to
+		 share signals.  Since we only want to report it once, we
+		 mark it as ignored for all LWPs except this one.  */
+	      iterate_over_lwps (set_ignore_sigint, NULL);
+	      lp->ignore_sigint = 0;
+	    }
+	  else
+	    maybe_clear_ignore_sigint (lp);
 	}
-      else
-	maybe_clear_ignore_sigint (lp);
     }
 
   /* This LWP is stopped now.  */